--=======================================================================================================
--  ROCK OBJECT SCRIPT
--
-- Author:		GIANTS (Bale.lua script)
-- Edit:		akuenzi (akuenzi@gmail.com)
-- Date:		11/23/2018
-- Purpose:		Creates a new 'rock' type of object.
--=======================================================================================================


Rock = {};

Rock.modDir = g_currentModDirectory
Rock.image = "Huds/rockIcon.dds"
Rock.background = "Huds/BGOverlay.dds"

local Rock_mt = Class(Rock, MountableObject);

InitObjectClass(Rock, "Rock");

function Rock:new(isServer, isClient, customMt)
    local mt = customMt;
    if mt == nil then
        mt = Rock_mt;
    end;

    local self = MountableObject:new(isServer, isClient, mt);

    self.forcedClipDistance = 300;
    registerObjectClassName(self, "Rock");

    self.fillType = FillUtil.FILLTYPE_ROCK;
    self.fillLevel = 0;
    self.rockValueScale = 1;

	g_currentMission:addLimitedObject(FSBaseMission.LIMITED_OBJECT_TYPE_ROCK, self);

    return self;
end;

function Rock:delete()
    if self.i3dFilename ~= nil then
        Utils.releaseSharedI3DFile(self.i3dFilename, nil, true);
    end
	
	if self.mapHotspot ~= nil then
		g_currentMission.ingameMap:deleteMapHotspot(self.mapHotspot);
	end;
	
    g_currentMission:removeLimitedObject(FSBaseMission.LIMITED_OBJECT_TYPE_ROCK, self);
    unregisterObjectClassName(self);
    g_currentMission:removeItemToSave(self);
    Rock:superClass().delete(self);
end;

function Rock:readStream(streamId, connection)
    local i3dFilename = Utils.convertFromNetworkFilename(streamReadString(streamId));
    if self.nodeId == 0 then
        self:createNode(i3dFilename);
    end;
    self.fillLevel = streamReadFloat32(streamId);
    Rock:superClass().readStream(self, streamId, connection);
    g_currentMission:addItemToSave(self);

    self.rockValueScale = streamReadFloat32(streamId);
end;

function Rock:writeStream(streamId, connection)
    streamWriteString(streamId, Utils.convertToNetworkFilename(self.i3dFilename));
    streamWriteFloat32(streamId, self.fillLevel);
    Rock:superClass().writeStream(self, streamId, connection);

    streamWriteFloat32(streamId, self.rockValueScale);
end;

function Rock:update(dt)
	Rock:superClass().update(self);
	if not self.firstRun then
		local _, y, _ = getWorldTranslation(self.nodeId)
		if y < -100 then
			self:delete()
		end;
		self.firstRun = true
	end;
	
	-- Check to see if the rock was newly spawned.  Set collision mask back to 'normal' after alotted time.
	if self.collisionUpdate then
		if self.collisionCounter > 0 then
			self.collisionCounter = self.collisionCounter - dt
			--print("Counter = " .. tostring(self.collisionCounter))
		else
			setCollisionMask(self.nodeId, 16789506)								-- Bits 1 & 13 added back.
			self.collisionUpdate, self.collisionCounter = nil, nil
		end;
	end;

end;

function Rock:updateTick(dt)
	Rock:superClass().updateTick(self, dt);
end;

function Rock:mount(object, node, x,y,z, rx,ry,rz)
    g_currentMission:removeItemToSave(self);
    Rock:superClass().mount(self, object, node, x,y,z, rx,ry,rz);
end;

function Rock:unmount()
    if Rock:superClass().unmount(self) then
        g_currentMission:addItemToSave(self);
        return true;
    end
    return false;
end;

function Rock:setNodeId(nodeId)
    Rock:superClass().setNodeId(self, nodeId);

    local defaultFillLevel = 100;
    self.fillLevel = defaultFillLevel
    self.rockValueScale = Utils.getNoNil(tonumber(getUserAttribute(nodeId, "rockValueScale")), 1);

    self.fillType = FillUtil.FILLTYPE_ROCK;
    local fillTypeStr = getUserAttribute(nodeId, "fillType");
    if fillTypeStr ~= nil then
        local fillTypeInt = FillUtil.fillTypeNameToInt[fillTypeStr];
        if fillTypeInt ~= nil then
            self.fillType = fillTypeInt;
        end
    end
	
	self.forcedClipDistance = 300
	
	--[[Add map hot spot for each rock so they can be found.  We put this here in the code, as it's a location that will end up
		getting called by load(), loadFromAttributesAndNodes(), and readStream().]]
	local width, height = getNormalizedScreenValues(4, 4)
	local objectId = self.nodeId
	local image = Utils.getFilename(Rock.image, Rock.modDir);
	local bgImage = Utils.getFilename(Rock.background, Rock.modDir);
	
	self.mapHotspot = g_currentMission.ingameMap:createMapHotspot("Rock", "", image, nil, nil, nil, nil, width, height, false, false, false, objectId, nil, MapHotspot.CATEGORY_DEFAULT, nil, nil, nil, bgImage, nil, Overlay.ALIGN_VERTICAL_MIDDLE, 1)
	-- Version below makes small gray circles on the map.
	--	self.mapHotspot = g_currentMission.ingameMap:createMapHotspot("Rock", "", nil, getNormalizedUVs({776, 776, 240, 240}), {0.5020, 0.5020, 0.5020, 1}, x, z, width, height, false, false, false, objectId, nil, MapHotspot.CATEGORY_DEFAULT, nil, nil, nil, nil, getNormalizedUVs({776, 776, 240, 240}), Overlay.ALIGN_VERTICAL_MIDDLE, 0.7)
	--function IngameMap:createMapHotspot(name, fullName, imageFilename, imageUVs, baseColor, xMapPos, zMapPos, width, height, blinking, persistent, showName, objectId, renderLast, category, textSize, textOffsetY, textColor, bgImageFilename, bgImageUVs, verticalAlignment, overlayBgScale)
	
end;

function Rock:createNode(i3dFilename)
    self.i3dFilename = i3dFilename;
    self.customEnvironment, self.baseDirectory = Utils.getModNameAndBaseDirectory(i3dFilename);
    local rockRoot = Utils.loadSharedI3DFile(i3dFilename);

    local rockId = getChildAt(rockRoot, 0);
    link(getRootNode(), rockId);
    delete(rockRoot);

    self:setNodeId(rockId);
end;

function Rock:load(i3dFilename, x,y,z, rx,ry,rz, fillLevel)
    self.i3dFilename = i3dFilename;
    self.customEnvironment, self.baseDirectory = Utils.getModNameAndBaseDirectory(i3dFilename);
    self:createNode(i3dFilename);
    setTranslation(self.nodeId, x, y, z);
    setRotation(self.nodeId, rx, ry, rz);
    if fillLevel ~= nil then
        self.fillLevel = fillLevel;
    end
	
	-- Code to set collision mask so that it temporarily won't interfere with the plough/cultivator.
	self.collisionCounter = 2000				-- 2 seconds
	self.collisionUpdate = true
	--print("Collision Mask = " .. tostring(getCollisionMask(self.nodeId)))
	setCollisionMask(self.nodeId, 16781312)		-- Bits 1 & 13 removed.
	
    g_currentMission:addItemToSave(self);
end;

function Rock:loadFromMemory(nodeId, i3dFilename)
    self.i3dFilename = i3dFilename;
    self.customEnvironment, self.baseDirectory = Utils.getModNameAndBaseDirectory(i3dFilename);
    self:setNodeId(nodeId);
end;

function Rock:loadFromAttributesAndNodes(xmlFile, key, resetVehicles)

    local x,y,z = Utils.getVectorFromString(getXMLString(xmlFile, key.."#position"));
    local xRot,yRot,zRot = Utils.getVectorFromString(getXMLString(xmlFile, key.."#rotation"));
    if x == nil or y == nil or z == nil or xRot == nil or yRot == nil or zRot == nil then
        return false;
    end;
    local filename = getXMLString(xmlFile, key.."#filename");
    if filename == nil then
        return false;
    end;
    filename = Utils.convertFromNetworkFilename(filename);
    local rootNode = Utils.loadSharedI3DFile(filename);
    if rootNode == 0 then
        return false;
    end;

    local ret = false;
    local node = getChildAt(rootNode, 0);
    if node ~= nil and node ~= 0 then
        setTranslation(node, x,y,z);
        setRotation(node, xRot,yRot,zRot);
        link(getRootNode(), node);
        ret = true;
    end;
    delete(rootNode);
    if not ret then
        return false;
    end;

    self:loadFromMemory(node, filename);

    local fillLevel = getXMLFloat(xmlFile, key.."#fillLevel");
    if fillLevel ~= nil then
        self.fillLevel = fillLevel;
    end;

    local rockValueScale = Utils.getNoNil(tonumber(getXMLString(xmlFile, key.."#rockValueScale")), 1);
    if rockValueScale ~= nil then
        self.rockValueScale = rockValueScale;
    end

    return true;
end;

function Rock:getSaveAttributesAndNodes(nodeIdent)

    local x,y,z = getTranslation(self.nodeId);
    local xRot,yRot,zRot = getRotation(self.nodeId);
	
	-- If rocks spawn below the map, they will be removed.
	if y <= -100 then
		return nil, nil
	end;

    local attributes = 'filename="'.. Utils.encodeToHTML(Utils.convertToNetworkFilename(self.i3dFilename))..'" position="'..x..' '..y..' '..z..'" rotation="'..xRot..' '..yRot..' '..zRot..'" rockValueScale="'..self.rockValueScale..'" fillLevel="'..self.fillLevel..'"'; --..wrapping;
    local nodes = "";
    return attributes, nodes;
end;

function Rock:getValue()
    local pricePerLiter = g_currentMission.economyManager:getPricePerLiter(self.fillType);
    return self.fillLevel * pricePerLiter * self.rockValueScale;
end;

function Rock:getFillType()
    return self.fillType;
end;

function Rock:getFillLevel()
    return self.fillLevel;
end;

function Rock:setFillLevel(fillLevel)
    self.fillLevel = fillLevel;
end;