--=======================================================================================================
--  ROCKLOAD SCRIPT
-- 
-- Author:		akuenzi (akuenzi@gmail.com)
-- Date:		12/30/2018
-- Purpose:		Allows trailers to autoload rocks.
--=======================================================================================================


RockLoad = {};
RockLoad.ModDir = g_currentModDirectory

--[[The script will vector search from a given point on the trailer to see what rocks are close to it.  Below are
the inner and outer distances from this point between which to search for the rocks.  We need an inner ring in
which to not search for rocks because the script would otherwise constantly see rocks inside the trailer itself,
and attempt picking them up again and again.  Below are default amounts for this.]]
RockLoad.SEARCH_RING_SIZE = 5			-- Band, in meters, away from trailer in which rocks are sought.
RockLoad.Y_OFFSET = 1					-- Meters above ground from which to look for rocks.
RockLoad.UNLOAD_OFFSET = 1.5			-- Meters outside of inner distance at which to dump rocks from trailer.

function RockLoad.prerequisitesPresent(specializations)
	return SpecializationUtil.hasSpecialization(Trailer, specializations)
end;

function RockLoad:load(savegame)
	
	self.setRockLoadStatus = SpecializationUtil.callSpecializationsFunction("setRockLoadStatus");
	self.setRockUnLoadStatus = SpecializationUtil.callSpecializationsFunction("setRockUnLoadStatus");
	self.getRockInRange = RockLoad.getRockInRange
	self.pickupRock = RockLoad.pickupRock
	self.unloadRock = RockLoad.unloadRock
	self.canIPickUpRocks = RockLoad.canIPickUpRocks
	self.drawCircle = RockLoad.drawCircle
	self.drawUnloadSpot = RockLoad.drawUnloadSpot

	self.rockLoadFirstRun = true
	self.rockLoadScriptCanActivate = false						-- Trailer meets (true) or does not meet (false) criteria for loading/unloading rock.
	self.rockLoadStatus = false									-- Whether rock loading is on (true) or off (false);
	self.rockUnLoadStatus = false								-- Whether rock unloading is on (true) or off (false);
	self.rockAreaShow = false									-- Whether rock working area is on (true) or off (false);
	self.tempTipShow = false									-- Whether temporary showing of dump area is on (true) or off (false);
	self.tempTipTimer = 3000									-- Amount of time to show the spot where rocks will dump.
	
	self.rockLoadSpeed = 500									-- Time in ms that each rock will load onto the trailer.  500 ms is 1/2 a second per rock.
	self.rockTimer = 0											-- Rock loading timer.
	self.searchDistanceInner = 0								-- To set default inner search ring.  Distance from exactFillRootNode.
	self.searchDistanceOuter = 0								-- To set default outer search ring.  Distance from exactFillRootNode.
	
end;

function RockLoad:delete()
end;

function RockLoad:readStream(streamId, connection)
	self:setRockLoadStatus(streamReadBool(streamId), true);
	self:setRockUnLoadStatus(streamReadBool(streamId), true);
end;

function RockLoad:writeStream(streamId, connection)
	streamWriteBool(streamId, self.rockLoadStatus);
	streamWriteBool(streamId, self.rockUnLoadStatus);
end;

function RockLoad:mouseEvent(posX, posY, isDown, isUp, button)
end;

function RockLoad:keyEvent(unicode, sym, modifier, isDown)
end;

function RockLoad:update(dt)

	if self.rockLoadFirstRun then
		-- The following also have 'trailer' specializations, but we do not want to use them to pick up rocks.
		if not (SpecializationUtil.hasSpecialization(ForageWagon, self.specializations) or
				SpecializationUtil.hasSpecialization(StrawBlower, self.specializations) or
				SpecializationUtil.hasSpecialization(MixerWagon, self.specializations) or
				SpecializationUtil.hasSpecialization(HookLiftTrailer, self.specializations) or
				SpecializationUtil.hasSpecialization(WaterTrailer, self.specializations) or
				SpecializationUtil.hasSpecialization(Overloading, self.specializations) or
				SpecializationUtil.hasSpecialization(Mower, self.specializations)) then
				
			-- Check to see if trailer has the needed 'spots' on trailer from which to drop rocks.
			if self.exactFillRootNode ~= nil and self.tipReferencePoints ~= nil then
			
				-- Points were found, so we can activate.
				self.rockLoadScriptCanActivate = true
				
				--[[Calculate placement of inner/outer rings for picking up of rocks.  The inner ring should be just slightly larger
					than the size of the trailer.  The exactFillRootNode appears to be in the middle of the trailer in most instances,
					so we'll calculate the distance from it to the furthest away tipReference to get the radius of a circle from the
					exactFillRootNode inside which rocks won't be picked up.]]
				local px, _, pz = localToWorld(self.exactFillRootNode, 0, 0, 0);
				
				-- Calculate maximum distance between exactFillRootNode and the trailer's tip points.
				local vector = 0
				for i = 1, table.getn(self.tipReferencePoints) do
					local a, _, b = localToWorld(self.tipReferencePoints[i].node, 0, 0, 0)
					local distX = a - px
					local distZ = b - pz
					vector = math.max(Utils.vector2Length(distX, distZ), vector)
				end;

				-- Calculate the distance at which rocks will be picked up based on approximate size of the trailer.
				local ringSize = RockLoad.SEARCH_RING_SIZE
				self.searchDistanceInner = vector
				self.searchDistanceOuter = self.searchDistanceInner + ringSize
			end;
		end;
		self.rockLoadFirstRun = false
	end;

	if self.rockLoadScriptCanActivate then
		if self:getIsActive() then
		
			-- Check to see if player turns loading on or off.
			if self:getIsActiveForInput() then
				if self:canIPickUpRocks() then
					-- Check to see if we are loading rocks.
					if InputBinding.hasEvent(InputBinding.RockLoadingStatus) then
						local statusChange = not self.rockLoadStatus
						self:setRockLoadStatus(statusChange);
						
						-- If we are set to load rocks while the trailer is unloading, then...
						if self.rockLoadStatus and self.rockUnLoadStatus then
							-- Turn OFF the unloading process.
							self:setRockUnLoadStatus(false);
						end;
					end;
					
					-- Check to see if we are unloading rocks.
					if InputBinding.hasEvent(InputBinding.RockUnLoadingStatus) then
						local statusChange = not self.rockUnLoadStatus
						self:setRockUnLoadStatus(statusChange);
						
						-- If we are set to unload rocks while the trailer is loading, then...
						if self.rockUnLoadStatus and self.rockLoadStatus then
							-- Turn OFF the loading process.
							self:setRockLoadStatus(false);
						end;
					end;
					
					-- Check to see if we are showing the rock pickup/unloading areas.
					if InputBinding.hasEvent(InputBinding.RockAreaShow) then
						self.rockAreaShow = not self.rockAreaShow
					end;
					
					-- If we are neither loading nor unloading and the rock area is 'showing,' then it should be turned off.
					if not self.rockLoadStatus and not self.rockUnLoadStatus and self.rockAreaShow then
						self.rockAreaShow = false
					end;
					
					-- If we toggle tip side so as to know where to unload, then we should temporarily show dump spots.
					if InputBinding.hasEvent(InputBinding.TOGGLE_TIPSIDE) then
						self.tempTipShow = true
						self.tempTipTimer = 3000
					end;
					
					if self.tempTipShow then
						self.tempTipTimer = self.tempTipTimer - dt
						if self.tempTipTimer <= 0 then
							self.tempTipShow = false
						end;
					end;			
					
				else
					self:setRockLoadStatus(false);
					self:setRockUnLoadStatus(false);
				end;
			end;
		else
			self:setRockLoadStatus(false);
			self:setRockUnLoadStatus(false);
		end;
	end;	
end;

function RockLoad:updateTick(dt)
	-- Perhaps we'll use updateTick since it's called less frequently than update.
	if self.rockLoadScriptCanActivate then
		-- If we're turned on to autoload rocks...
		if self.rockLoadStatus then
			self.rockTimer = self.rockTimer + dt
			-- Are we ready to pick up a rock?
			if self.rockTimer >= self.rockLoadSpeed then
				-- Reset timer
				self.rockTimer = 0
				
				-- Find the rock closest to the trailer.
				local nearestRock = self:getRockInRange(true)
				if nearestRock ~= nil then
					-- Are we the server (where the rocks are stored)?
					if self.isServer then
						self:pickupRock(nearestRock);
					end;
				end;
			end;
		end;
		
		-- If we're turned on to auto unload rocks...
		if self.rockUnLoadStatus then
			self.rockTimer = self.rockTimer + dt
			-- Are we ready to unload a rock?
			if self.rockTimer >= self.rockLoadSpeed then
				-- Reset timer
				self.rockTimer = 0
				
				-- Find a rock to unload from the trailer
				local nearestRock = self:getRockInRange(false)
				if nearestRock ~= nil then
					-- Are we the server (where the rocks are stored)?
					if self.isServer then
						self:unloadRock(nearestRock);
					end;
				end;
			end;
		end;
	end;
end;

function RockLoad:draw()
	-- To generate message to appear in the F1 help menu.
	if self.rockLoadScriptCanActivate and self.isClient then
		if self:getIsActiveForInput(true) then
			if self:canIPickUpRocks() then
				-- Help text for rock loading status.
				if self.rockLoadStatus then
					g_currentMission:addHelpButtonText(string.format(g_i18n:getText("ROCK_LOAD_STATUS_TURN_OFF"), g_i18n:getText("input_RockLoadingStatus")), InputBinding.RockLoadingStatus, nil, GS_PRIO_HIGH);
					
					if self.rockAreaShow then
						-- Help Text.  This button won't even appear unless we've activated either loading or unloading, so as to save space on the help menu.
						g_currentMission:addHelpButtonText(string.format(g_i18n:getText("ROCK_LOAD_STATUS_TURN_OFF"), g_i18n:getText("input_RockAreaShow")), InputBinding.RockAreaShow, nil, GS_PRIO_HIGH);
						
						-- Draw pick up area.
						local x, _, z = getWorldTranslation(self.exactFillRootNode) -- Center of circle.
						local radiusInner = self.searchDistanceInner
						local radiusOuter = self.searchDistanceOuter
						-- Calculate circumference in meters, which will give the number of steps.
						local stepsInner = Utils.round(math.pi * (radiusInner * 2))
						local stepsOuter = Utils.round(math.pi * (radiusOuter * 2))
						self:drawCircle(x, z, radiusInner, stepsInner)
						self:drawCircle(x, z, radiusOuter, stepsOuter)
					else
						-- Help Text.
						g_currentMission:addHelpButtonText(string.format(g_i18n:getText("ROCK_LOAD_STATUS_TURN_ON"), g_i18n:getText("input_RockAreaShow")), InputBinding.RockAreaShow, nil, GS_PRIO_HIGH);
					end;
				else
					g_currentMission:addHelpButtonText(string.format(g_i18n:getText("ROCK_LOAD_STATUS_TURN_ON"), g_i18n:getText("input_RockLoadingStatus")), InputBinding.RockLoadingStatus, nil, GS_PRIO_HIGH);
				end;
				
				-- Help text for rock unloading status.
				if self.rockUnLoadStatus then
					g_currentMission:addHelpButtonText(string.format(g_i18n:getText("ROCK_LOAD_STATUS_TURN_OFF"), g_i18n:getText("input_RockUnLoadingStatus")), InputBinding.RockUnLoadingStatus, nil, GS_PRIO_HIGH);
				
					-- Draw zone in which rocks are pulled for drop and the spot where they'll be dropped.  Won't show until there are rocks to send out.
					if self.rockAreaShow then
						-- Help Text.
						g_currentMission:addHelpButtonText(string.format(g_i18n:getText("ROCK_LOAD_STATUS_TURN_OFF"), g_i18n:getText("input_RockAreaShow")), InputBinding.RockAreaShow, nil, GS_PRIO_HIGH);
						self:drawUnloadSpot() 
					else
						-- Help Text.
						g_currentMission:addHelpButtonText(string.format(g_i18n:getText("ROCK_LOAD_STATUS_TURN_ON"), g_i18n:getText("input_RockAreaShow")), InputBinding.RockAreaShow, nil, GS_PRIO_HIGH);
					end;
				else
					g_currentMission:addHelpButtonText(string.format(g_i18n:getText("ROCK_LOAD_STATUS_TURN_ON"), g_i18n:getText("input_RockUnLoadingStatus")), InputBinding.RockUnLoadingStatus, nil, GS_PRIO_HIGH);
				end;
				
				-- Rendering of dump areas when tip area changes.
				if self.tempTipShow then
					self:drawUnloadSpot()
				end;			
			end;
		end;	
	end;	
end;

function RockLoad:setRockLoadStatus(RockLoadStatus, noEventSend)
	-- Update trailer loading status for input binding changes and sync in a multiplayer session.
	if self.rockLoadStatus ~= RockLoadStatus then
		if noEventSend == nil or noEventSend == false then
			if g_server ~= nil then
				g_server:broadcastEvent(setRockLoadStatusEvent:new(self, RockLoadStatus), nil, nil, self);
			else
				g_client:getServerConnection():sendEvent(setRockLoadStatusEvent:new(self, RockLoadStatus));
			end;
		end;
		
		self.rockLoadStatus = RockLoadStatus;
	end;
end;

function RockLoad:setRockUnLoadStatus(RockUnLoadStatus, noEventSend)
	-- Update trailer unloading status for input binding changes and sync in a multiplayer session.
	if self.rockUnLoadStatus ~= RockUnLoadStatus then
		if noEventSend == nil or noEventSend == false then
			if g_server ~= nil then
				g_server:broadcastEvent(setRockUnLoadStatusEvent:new(self, RockUnLoadStatus), nil, nil, self);
			else
				g_client:getServerConnection():sendEvent(setRockUnLoadStatusEvent:new(self, RockUnLoadStatus));
			end;
		end;
		
		self.rockUnLoadStatus = RockUnLoadStatus;
	end;
end;

function RockLoad:getRockInRange(loading)		-- A value of 'true' means we are loading rocks.  A value of 'false' means we are unloading rocks.

	-- Get world position of the trailer's exactFillRootNode.
	local px, py, pz = getWorldTranslation(self.exactFillRootNode);
	-- Get ground height of exactFillRootNode (ie, below the trailer).
	local terrain = g_currentMission.terrainRootNode
	local groundHeight = getTerrainHeightAtWorldPos(terrain, px, py, pz) + RockLoad.Y_OFFSET;
	-- Get range (inner and outer rings) in which to look for rocks.
	local nearestDistance = self.searchDistanceOuter;
	local awayDistance = self.searchDistanceInner;
	-- Get our starter rock, which is 'nothing.'
	local nearestRock = nil;
	
	-- Parse save items table to find rocks.
	for _, item in pairs(g_currentMission.itemsToSave) do
		local rock = item.item;
		-- If the item is a rock, then...
		if rock:isa(Rock) then
			-- Find the location of the rock on the map.
			local vx, vy, vz = getWorldTranslation(rock.nodeId);
			-- Calculate the distance of the rock to our trailer point.
			local distance = Utils.vector3Length(px-vx, groundHeight-vy, pz-vz);
			-- If the rock is in range, then...
			if loading then			-- We are loading a rock.  Find rocks outside of trailer to load.
				if distance < nearestDistance and distance > awayDistance then
					nearestDistance = distance;
					nearestRock = rock.nodeId;
				end;
			else					-- We are unloading a rock.  Find rocks inside trailer to unload.
				if distance < awayDistance then
					awayDistance = distance;
					nearestRock = rock.nodeId;
				end;
			end;
		end;
	end;
	
	-- Send back the identity of the closest rock.
	return nearestRock;
end;

function RockLoad:pickupRock(rockId)

	-- Get world position of the exactFillRootNode (place where we want the rock to drop into the trailer).
	local px, py, pz = getWorldTranslation(self.exactFillRootNode);
	
	-- Move the rock from its current location to the exactFillRootNode.
	removeFromPhysics(rockId);
	setTranslation(rockId, px, py, pz);
	addToPhysics(rockId);
	
end;

function RockLoad:unloadRock(rockId)

	--[[Get world position of exactFillRootNode, which is central in the trailer and the point from which rocks 
		are searched out.]]
	local px, py, pz = getWorldTranslation(self.exactFillRootNode);
	
	--[[Get world position of the current trailer tipReferencePoint, which is approximately the place where
		we want to drop the rock out of the trailer).]]
	local a, _, b = getWorldTranslation(self.tipReferencePoints[self.preferedTipReferencePointIndex].node)
	
	--[[Get distance between the above two points on the map so as to know what direction/location we are to 
		move the rocks.]]
	local distX = a - px
	local distZ = b - pz
	local vector = Utils.vector2Length(distX, distZ)
	
	--[[We key the rock movement from the exactFillRootNode using the distance and direction given from the 
		points above, and multiply by a distance factor to ensure the rock drops clear of the edge of the
		trailer.  The inner ring is used, as this is the outer bound from which rocks are removed from the trailer.]]
	local offsetRatio = (self.searchDistanceInner + RockLoad.UNLOAD_OFFSET) / vector
	
	px = px + distX * offsetRatio
	pz = pz + distZ * offsetRatio
	
	-- Move the rock from its current location in the trailer to rock drop point.
	removeFromPhysics(rockId);
	setTranslation(rockId, px, py, pz);
	addToPhysics(rockId);

end;

function RockLoad:canIPickUpRocks()
	--[[Picking up rocks can be resource intensive, and it's probably not a good idea to leave
		the trailer auto-picking up rocks while it's not attended or otherwise filled with something.]]
	
	-- Is the trailer attached?
	if self.attacherVehicle == nil then
		return false
	end;
	
	-- Is the trailer in a 'tipped' or 'tipping' state?
	if self.tipState ~= nil and self.tipState ~= Trailer.TIPSTATE_CLOSED then
		return false
	end;
	
	-- Does the trailer already have material inside it?
	if self.fillUnits ~= nil then
		local fillUnits = self.fillUnits
		for _, unit in pairs(fillUnits) do
			local fillLevel = Utils.getNoNil(unit.fillLevel, 0)
			if fillLevel > 0 then
				return false
			end;
		end;
	end;
	
	-- Is the trailer covered?
	if not self.isCoverOpen then
		return false
	end;
	
	return true
end;

function RockLoad:drawCircle(x, z, radius, steps)

	local terrain = g_currentMission.terrainRootNode;
	for i = 1, steps do
		local a1 = ((i-1)/steps)*2*math.pi;
		local a2 = ((i)/steps)*2*math.pi;

		local c = math.cos(a1) * radius;
		local s = math.sin(a1) * radius;
		local x1, z1 = x+c, z+s;

		local c = math.cos(a2) * radius;
		local s = math.sin(a2) * radius;
		local x2, z2 = x+c, z+s;

		local y1 = getTerrainHeightAtWorldPos(terrain, x1, 0, z1) + RockLoad.Y_OFFSET;	
		local y2 = getTerrainHeightAtWorldPos(terrain, x2, 0, z2) + RockLoad.Y_OFFSET;

		drawDebugLine(x1,y1,z1, 0.25,0.25,0.25, x2,y2,z2, 0.25,0.25,0.25);
	end;
end;

function RockLoad:drawUnloadSpot()

	local px, py, pz = getWorldTranslation(self.exactFillRootNode);
	local a, _, b = getWorldTranslation(self.tipReferencePoints[self.preferedTipReferencePointIndex].node)
	local distX = a - px
	local distZ = b - pz
	local vector = Utils.vector2Length(distX, distZ)
	local offsetRatio = (self.searchDistanceInner + RockLoad.UNLOAD_OFFSET) / vector
	px = px + distX * offsetRatio
	pz = pz + distZ * offsetRatio
	local radiusDrop = 1
	local stepsDrop = Utils.round(math.pi * (radiusDrop * 2))
	self:drawCircle(px, pz, radiusDrop, stepsDrop)

end;


---------------------------------------------------------------------------------------------------------
--  SET ROCK LOADING STATUS EVENT FOR MULTIPLAYER	(ie, WHEN TRAILER LOADING INPUT BINDING CHANGES)
---------------------------------------------------------------------------------------------------------
setRockLoadStatusEvent = {};
setRockLoadStatusEvent_mt = Class(setRockLoadStatusEvent, Event);

InitEventClass(setRockLoadStatusEvent, "setRockLoadStatusEvent");

function  setRockLoadStatusEvent:emptyNew()
	local self = Event:new(setRockLoadStatusEvent_mt);
	self.className = "setRockLoadStatusEvent";
	return self;
end;

function setRockLoadStatusEvent:new(object,RockLoadStatus)
	local self = setRockLoadStatusEvent:emptyNew()
	self.object = object;
	self.rockLoadStatus = RockLoadStatus;
	return self;
end;

function setRockLoadStatusEvent:readStream(streamId, connection)
	self.object = networkGetObject(streamReadInt32(streamId));
	self.rockLoadStatus = streamReadBool(streamId);
	self:run(connection);
end;

function setRockLoadStatusEvent:writeStream(streamId, connection)
	streamWriteInt32(streamId, networkGetObjectId(self.object));
	streamWriteBool(streamId,self.rockLoadStatus);
end;

function setRockLoadStatusEvent:run(connection)
	self.object:setRockLoadStatus(self.rockLoadStatus, true);
	if not connection:getIsServer() then
		g_server:broadcastEvent(setRockLoadStatusEvent:new(self.object,self.rockLoadStatus),nil,connection,self.object);
	end;
end;


---------------------------------------------------------------------------------------------------------
--  SET ROCK UNLOADING STATUS EVENT FOR MULTIPLAYER	(ie, WHEN TRAILER UNLOADING INPUT BINDING CHANGES)
---------------------------------------------------------------------------------------------------------
setRockUnLoadStatusEvent = {};
setRockUnLoadStatusEvent_mt = Class(setRockUnLoadStatusEvent, Event);

InitEventClass(setRockUnLoadStatusEvent, "setRockUnLoadStatusEvent");

function  setRockUnLoadStatusEvent:emptyNew()
	local self = Event:new(setRockUnLoadStatusEvent_mt);
	self.className = "setRockUnLoadStatusEvent";
	return self;
end;

function setRockUnLoadStatusEvent:new(object, RockUnLoadStatus)
	local self = setRockUnLoadStatusEvent:emptyNew()
	self.object = object;
	self.rockUnLoadStatus = RockUnLoadStatus;
	return self;
end;

function setRockUnLoadStatusEvent:readStream(streamId, connection)
	self.object = networkGetObject(streamReadInt32(streamId));
	self.rockUnLoadStatus = streamReadBool(streamId);
	self:run(connection);
end;

function setRockUnLoadStatusEvent:writeStream(streamId, connection)
	streamWriteInt32(streamId, networkGetObjectId(self.object));
	streamWriteBool(streamId,self.rockUnLoadStatus);
end;

function setRockUnLoadStatusEvent:run(connection)
	self.object:setRockUnLoadStatus(self.rockUnLoadStatus, true);
	if not connection:getIsServer() then
		g_server:broadcastEvent(setRockUnLoadStatusEvent:new(self.object,self.rockUnLoadStatus),nil,connection,self.object);
	end;
end;