--=======================================================================================================
--  ROCKSPAWN SCRIPT
-- 
-- Author:		akuenzi (akuenzi@gmail.com)
-- Date:		12/2/2018 - Update to FS '17
--				6/28/2016 - Original FS '15
-- Purpose:		Adds rock spawning capability to cultivators/plows
--=======================================================================================================


source(g_currentModDirectory .. "Scripts/Rock.lua");

RockSpawn = {};
RockSpawn.ModDir = g_currentModDirectory
RockSpawn.Debug = false

-- Rock i3d files.
RockSpawn.RockFiles = {
	{"Rocks/Giants_FieldRock01.i3d", fillLevel = 370},
	{"Rocks/Giants_FieldRock02.i3d", fillLevel = 280},
	{"Rocks/Giants_FieldRock03.i3d", fillLevel = 210},
	{"Rocks/Giants_rock1.i3d", fillLevel = 60},
	{"Rocks/Giants_17stone01.i3d", fillLevel = 60},
	{"Rocks/Giants_17stone02.i3d", fillLevel = 60},
	{"Rocks/Giants_17stone03.i3d", fillLevel = 60},
	{"Rocks/Giants_17stone04.i3d", fillLevel = 90},
	{"Rocks/Giants_17stone05.i3d", fillLevel = 60},
	{"Rocks/Giants_17stone07.i3d", fillLevel = 60},
	{"Rocks/Giants_17stone08.i3d", fillLevel = 60},
	{"Rocks/TyroSmith_Rock1.i3d", fillLevel = 40},
	{"Rocks/TyroSmith_Rock2.i3d", fillLevel = 30},
	{"Rocks/TyroSmith_Rock3.i3d", fillLevel = 90},
	{"Rocks/TyroSmith_Rock4.i3d", fillLevel = 50},
	{"Rocks/TyroSmith_Rock5.i3d", fillLevel = 50},
	{"Rocks/TyroSmith_Rock6.i3d", fillLevel = 140},
	{"Rocks/Benw11_Rock1.i3d", fillLevel = 90},
	{"Rocks/Benw11_Rock2.i3d", fillLevel = 40},	
	{"Rocks/Benw11_Rock3.i3d", fillLevel = 20},	
	{"Rocks/Benw11_Rock4.i3d", fillLevel = 10},	
	{"Rocks/Giants_FieldRock01g.i3d", fillLevel = 0.4},										-- This is the 'special' rock (gold).  It should always be LAST in this list.		
};

RockSpawn.MAX_NUM_ROCKS = table.getn(RockSpawn.RockFiles) - 1;

-- Sound effect files for impact of rocks with tillage equipment.  Mono wav files with 44100 sample rate.
RockSpawn.RockSounds = {
	"Sounds/impact_heavy_stone_land_on_ground_001.wav",
	"Sounds/impact_heavy_stone_land_on_ground_002.wav",
	"Sounds/impact_metal_water_trough.wav",
	"Sounds/industrial_tools_shovel_metal_hit_concrete_floor_001.wav",
	"Sounds/industrial_tools_shovel_metal_hit_concrete_floor_002.wav",
	"Sounds/industrial_tools_shovel_metal_hit_concrete_floor_003.wav",
	"Sounds/zapsplat_foley_heavy_rock_drop_on_ground_001.wav",
	"Sounds/zapsplat_foley_heavy_rock_drop_on_ground_003.wav",
	"Sounds/zapsplat_household_rake_metal_drag_on_concrete_short_001_17810.wav",
	"Sounds/zapsplat_household_rake_metal_set_down_on_concrete_001_17812.wav",
	"Sounds/zapsplat_impact_metal_drain_stone_002_10962.wav",
	"Sounds/zapsplat_impact_metal_wall_fist_001_19348.wav",
	"Sounds/zapsplat_impacts_metal_hollow_cabinet_old_knock_16670.wav",
	"Sounds/zapsplat_industrial_hammer_hit_metal_x1+_12828.wav",
	"Sounds/zapsplat_industrial_metal_rake_hit_ground_001_11849.wav",
	"Sounds/zapsplat_industrial_metal_rake_hit_ground_002_11850.wav",
	"Sounds/zapsplat_industrial_shovel_metal_dig_dirt_soil_001_13540.wav",
	"Sounds/zapsplat_industrial_step_ladder_metal_step_hit_foot_hard_004_14211.wav",
}

RockSpawn.MAX_NUM_SOUNDS = table.getn(RockSpawn.RockSounds)

-- Input Binding for Rock Spawning Levels
RockSpawn.RockLevels = {
	{rlevel = 0, rMinDistance = 0, rMaxDistance = 0},										-- The input binding of '0' is intended to represent the 'off' position.
	{rlevel = 1, rMinDistance = 284.4, rMaxDistance = 853.1},								-- rMinDistance = minimum distance after last rock spawn, that another rock may spawn.
	{rlevel = 2, rMinDistance = 142.2, rMaxDistance = 426.5},								-- rMaxDistance = maximum distance after last rock spawn, that another rock may spawn.
	{rlevel = 3, rMinDistance = 71.1, rMaxDistance = 213.3},
	{rlevel = 4, rMinDistance = 35.5, rMaxDistance = 106.6},
	{rlevel = 5, rMinDistance = 17.8, rMaxDistance = 53.3},	
};

RockSpawn.MAX_LEVELS = table.getn(RockSpawn.RockLevels) - 1;

-- Rock spawn distances
RockSpawn.DEFAULT_WIDTH = 12;																-- Default width of plow/cultivator.  Larger width = less spawning overall.

-- General offset values
RockSpawn.Y_OFFSET = 0.15;																	-- Move spawn point up 

function RockSpawn.prerequisitesPresent(specializations)
	return (SpecializationUtil.hasSpecialization(Cultivator, specializations) or SpecializationUtil.hasSpecialization(Plough, specializations)) and SpecializationUtil.hasSpecialization(WorkArea, specializations)
end;

function RockSpawn:load(savegame)
	self.setRockSpawnStatus = SpecializationUtil.callSpecializationsFunction("setRockSpawnStatus");	
	self.createRock = RockSpawn.createRock;
	self.randomSpawnPoint = RockSpawn.randomSpawnPoint
	
	-- Determine the width of the plow/cultivator.
	local storeItem = StoreItemsUtil.storeItemsByXMLFilename[self.configFileName:lower()]
	local width = storeItem.specs.workingWidth
	if width ~= nil then
		self.implementWidth = width
	elseif self.aiLeftMarker ~= nil and self.aiRightMarker ~= nil then
		local a, _, c = getTranslation(self.aiLeftMarker);
		local d, _, e = getTranslation(self.aiRightMarker);
		self.implementWidth = math.max(math.abs(d - a), math.abs(e - c))
	else
		self.implementWidth = RockSpawn.DEFAULT_WIDTH
	end;
	
	-- Determine if the script can activate or not.
	if self.workAreas[1].start ~= nil and self.workAreas[1].width ~= nil and self.workAreas[1].height ~= nil then
		if self.implementWidth ~= nil then
			self.scriptCanActivate = true
		else
			print("RockSpawn will not work with this file - [[[",self.configFileName,"]]].  It is missing aiMarkers or store width specifications.")
		end;
	else
		print("RockSpawn will not work with this file - [[[",self.configFileName,"]]].  It is missing at least one properly set up work area.")
	end;
	
	-- To determine 'rock areas' (by implement type) that will be used later in getting the work area coordinates.
	if SpecializationUtil.hasSpecialization(Plough, self.specializations) then
		self.rockAreaType = WorkArea.AREATYPE_PLOUGH
	else
		self.rockAreaType = WorkArea.AREATYPE_CULTIVATOR
	end;
	
	-- To set initial rock digging status of the plow/cultivator.  [2] is level 1.
	self.RockSpawnStatus = RockSpawn.RockLevels[2]["rlevel"];																-- Default rock spawning level at start:  Level 1

	-- To load rock spawning sounds.
	self.playSoundStatus = SpecializationUtil.callSpecializationsFunction("playSoundStatus");
	self.playRockSound = RockSpawn.playRockSound
	
	if self.isClient then
		self.sampleRockSound = {}
		for i = 1, RockSpawn.MAX_NUM_SOUNDS do
			self.sampleRockSound[i] = createSample("sampleRockSound");
			loadSample(self.sampleRockSound[i], RockSpawn.ModDir .. RockSpawn.RockSounds[i], true)
		end;
		self.rockSoundToPlay = self.sampleRockSound[13]																		-- To set initial rock sound to play as a temporary placeholder.
		--print("Initial Rock Sound to Play: " .. tostring(self.rockSoundToPlay))
	end;
	
end;

function RockSpawn:postLoad(savegame)	
	if savegame ~= nil and not savegame.resetVehicles then
		local status = getXMLInt(savegame.xmlFile, savegame.key.."#RockSpawnStatus");
		if status ~= nil then
			self:setRockSpawnStatus(status);
		end;
	end;
	RockSpawn.resetDistance(self);
end;

function RockSpawn:delete()
	if self.isClient then
		for i = 1, RockSpawn.MAX_NUM_SOUNDS do
			delete(self.sampleRockSound[i]);
		end;
	end;
end;

function RockSpawn:readStream(streamId, connection)
	self:setRockSpawnStatus(streamReadInt8(streamId), true);
	self.rockSoundToPlay = streamReadInt32(streamId);
end;

function RockSpawn:writeStream(streamId, connection)
	streamWriteInt8(streamId,self.RockSpawnStatus);
	streamWriteInt32(streamId, self.rockSoundToPlay);
end;

function RockSpawn:getSaveAttributesAndNodes(nodeIdent)
	local attributes = 'RockSpawnStatus="' .. self.RockSpawnStatus .. '"';
	return attributes, nil;
end;

function RockSpawn:mouseEvent(posX, posY, isDown, isUp, button)
end;

function RockSpawn:keyEvent(unicode, sym, modifier, isDown)
end;

function RockSpawn:update(dt)
	if self.scriptCanActivate then
		if self:getIsActive() then
		
			-- Check to see if player changes level of rock generation.
			if self:getIsActiveForInput() then
				if InputBinding.hasEvent(InputBinding.RockStatus) then
					local statusChange = self.RockSpawnStatus
					if statusChange == RockSpawn.MAX_LEVELS then
						statusChange = 0
					else
						statusChange = self.RockSpawnStatus + 1
					end;
					
					self:setRockSpawnStatus(statusChange);
					RockSpawn.resetDistance(self);
				end;
			end;
			
			-- Check to see if we've traveled the alotted distance required to generate next rock.
			if self.isServer and self.RockSpawnStatus ~= 0 then
				if self.doGroundManipulation or self.ploughHasGroundContact then
					if not self.showFieldNotOwnedWarning then
						if self.traveledDistance >= self.currentSpawnTargetDistance then
							RockSpawn.resetDistance(self);
							self:createRock();
							self:playSoundStatus(0);									-- A play status of '0' means a new rock has spawned and that a new sound needs to be selected.
							--print("Rock spawned, rock sound to play is: " .. tostring(self.rockSoundToPlay))
						else
							self.traveledDistance = self.traveledDistance + self.lastMovedDistance;
						end;
					end;
				end;
			end;
			
			-- Play sound when rock is spawned.
			if self.isClient and self.RockSpawnStatus ~= 0 then
				-- If a toggle of '0' is sent from the server, then generate a sound.
				if self.rockSoundToPlay == 0 then
					local i = math.random(1, RockSpawn.MAX_NUM_SOUNDS)
					self.rockSoundToPlay = self.sampleRockSound[i]						-- The actual sound to play upon rock spawning, which changes it away from 0 and thus will happen only once.
					self:playRockSound(true)
					--print("The sound has played.  My new rock sound is: " .. tostring(self.rockSoundToPlay))
				end;
			end;
		end;
	end;	
end;

function RockSpawn:updateTick(dt)

end;

function RockSpawn:draw()
	-- To generate message to appear in the F1 help menu.
	if self.scriptCanActivate then
		if self:getIsActiveForInput(true) then
			if self.RockSpawnStatus == 0 then
				g_currentMission:addHelpButtonText(string.format(g_i18n:getText("LEVEL_0"), g_i18n:getText("input_RockStatus")), InputBinding.RockStatus, nil, GS_PRIO_HIGH);
			else
				local helpText
				if self.RockSpawnStatus == 1 then
					helpText = g_i18n:getText("LEVEL_2")
				elseif self.RockSpawnStatus == 3 then
					helpText = g_i18n:getText("LEVEL_3")
				elseif self.RockSpawnStatus == 5 then
					helpText = g_i18n:getText("LEVEL_4")
				else
					helpText = ""
				end;
							
				g_currentMission:addHelpButtonText(string.format(g_i18n:getText("WE_ARE_ACTIVE"), g_i18n:getText("input_RockStatus"), g_i18n:getText("LEVEL_1"), RockSpawn.RockLevels[self.RockSpawnStatus+1]["rlevel"], helpText), InputBinding.RockStatus, nil, GS_PRIO_HIGH);
			end;
		end;	
	end;	
end;

function RockSpawn:createRock()

	-- To make sure this all takes place on the server vs. client.
	assert(self.isServer ~= nil);
	
	-- To choose the next rock to appear when spawning occurs.
	local rockId = math.ceil(math.random(1, RockSpawn.MAX_NUM_ROCKS * 10 + 1) / 10);
	
	-- This gets the map location of the work area that is currently being worked.
	local workAreas, _, _ = self:getTypedNetworkAreas(self.rockAreaType, true);
	
	-- This chooses a random point within a specific workArea in which to spawn a rock.
	local numAreas = table.getn(workAreas);
	local areaChosen = math.random(1, numAreas)
	local xSpawn, zSpawn = self:randomSpawnPoint(workAreas[areaChosen])
	local ySpawn = getTerrainHeightAtWorldPos(g_currentMission.terrainRootNode, xSpawn, 0, zSpawn) + RockSpawn.Y_OFFSET;
	
	-- Spawn a rock with the information above.
	if not RockSpawn.Debug then
		local rock = Rock:new(self.isServer, self.isClient);
		rock:load(Utils.getFilename(RockSpawn.RockFiles[rockId][1], RockSpawn.ModDir), xSpawn, ySpawn, zSpawn, math.random(0,360), math.random(0,360), math.random(0,360), RockSpawn.RockFiles[rockId]["fillLevel"]);
		rock:register();
	else											-- Debugging.  This should generate every single rock in the set all in a row.
		for i = 1, (RockSpawn.MAX_NUM_ROCKS + 1 ) do
			local rock = Rock:new(self.isServer, self.isClient);
			rock:load(Utils.getFilename(RockSpawn.RockFiles[i][1], RockSpawn.ModDir), (xSpawn + 20), (ySpawn + 3), (zSpawn - i * 2), math.random(0,360), math.random(0,360), math.random(0,360), RockSpawn.RockFiles[i]["fillLevel"]);
			rock:register();	
		end;
	end;
end;

function RockSpawn:randomSpawnPoint(workArea)

	-- Load world work areas:  start index (x,z), width index (x1,z1), and height index (x2,z2)
	local x = workArea[1]
	local z = workArea[2]
	local x1 = workArea[3]
	local z1 = workArea[4]
	local x2 = workArea[5]
	local z2 = workArea[6]
	
	-- Generate random numbers between 0 and 1.
	local a, b = math.random(), math.random()
	
	-- Generate random position x, z spawn coordinates within the workArea parallelogram.
	local xSpawn = x + a * (x1 - x) + b * (x2 - x)
	local zSpawn = z + a * (z1 - z) + b * (z2 - z)
	
	return xSpawn, zSpawn
end;

function RockSpawn:setRockSpawnStatus(RockSpawnStatus, noEventSend)
	-- Update spawn status for input binding changes and sync in a multiplayer session.
	if self.RockSpawnStatus ~= RockSpawnStatus then
		if noEventSend == nil or noEventSend == false then
			if g_server ~= nil then
				g_server:broadcastEvent(setRockSpawnStatusEvent:new(self, RockSpawnStatus), nil, nil, self);
			else
				g_client:getServerConnection():sendEvent(setRockSpawnStatusEvent:new(self, RockSpawnStatus));
			end;
		end;
		
		self.RockSpawnStatus = RockSpawnStatus;
	end;
end;

function RockSpawn:resetDistance()
	self.traveledDistance = 0;
	self.currentSpawnTargetDistance = math.random(RockSpawn.RockLevels[self.RockSpawnStatus+1]["rMinDistance"]*(RockSpawn.DEFAULT_WIDTH/self.implementWidth), RockSpawn.RockLevels[self.RockSpawnStatus+1]["rMaxDistance"]*(RockSpawn.DEFAULT_WIDTH/self.implementWidth));
end;

function RockSpawn:playSoundStatus(sound, noEventSend)

	if self.rockSoundToPlay ~= sound then
		if noEventSend == nil or noEventSend == false then
			if g_server ~= nil then
				g_server:broadcastEvent(rockSpawnSoundEvent:new(self, sound), nil, nil, self);
			else
				g_client:getServerConnection():sendEvent(rockSpawnSoundEvent:new(self, sound));
			end;
		end;
		
		self.rockSoundToPlay = sound
	end;
end;

function RockSpawn:playRockSound(isPlaying)
	if self.rockSoundToPlay ~= 0 and self:getIsActiveForSound() then
		if isPlaying then
			playSample(self.rockSoundToPlay, 1, 1, 0);
		else
			--RockSpawn.onDeactivateSounds(self);
		end;
	end;
end;

--[[function RockSpawn:onDeactivateSounds()
	if self.isClient and self.rockSoundToPlay ~= 0 then
		stopSample(self.rockSoundToPlay);
	end;
end;]]


---------------------------------------------------------------------------------------------------------
--  SET ROCK SPAWN STATUS EVENT FOR MULTIPLAYER	(ie, WHEN PLOW/CULTIVATOR SPAWNING INPUT BINDING CHANGES)
---------------------------------------------------------------------------------------------------------
setRockSpawnStatusEvent = {};
setRockSpawnStatusEvent_mt = Class(setRockSpawnStatusEvent, Event);

InitEventClass(setRockSpawnStatusEvent, "setRockSpawnStatusEvent");

function  setRockSpawnStatusEvent:emptyNew()
	local self = Event:new(setRockSpawnStatusEvent_mt);
	self.className = "setRockSpawnStatusEvent";
	return self;
end;

function setRockSpawnStatusEvent:new(object,RockSpawnStatus)
	local self = setRockSpawnStatusEvent:emptyNew()
	self.object = object;
	self.RockSpawnStatus = RockSpawnStatus;
	return self;
end;

function setRockSpawnStatusEvent:readStream(streamId, connection)
	self.object = networkGetObject(streamReadInt32(streamId));
	self.RockSpawnStatus = streamReadInt8(streamId);
	self:run(connection);
end;

function setRockSpawnStatusEvent:writeStream(streamId, connection)
	streamWriteInt32(streamId, networkGetObjectId(self.object));
	streamWriteInt8(streamId,self.RockSpawnStatus);
end;

function setRockSpawnStatusEvent:run(connection)
	self.object:setRockSpawnStatus(self.RockSpawnStatus, true);
	if not connection:getIsServer() then
		g_server:broadcastEvent(setRockSpawnStatusEvent:new(self.object,self.RockSpawnStatus),nil,connection,self.object);
	end;
end;


---------------------------------------------------------------------------------------------------------
--  SET ROCK SOUND EVENT FOR MULTIPLAYER	(ie, WHEN ROCK SPAWNS, LET CLIENT KNOW TO PLAY A SOUND)
---------------------------------------------------------------------------------------------------------
rockSpawnSoundEvent = {};
rockSpawnSoundEvent_mt = Class(rockSpawnSoundEvent, Event);

InitEventClass(rockSpawnSoundEvent, "rockSpawnSoundEvent");

function  rockSpawnSoundEvent:emptyNew()
	local self = Event:new(rockSpawnSoundEvent_mt);
	self.className = "rockSpawnSoundEvent";
	return self;
end;

function rockSpawnSoundEvent:new(object, sound)
	local self = rockSpawnSoundEvent:emptyNew()
	self.object = object;
	self.rockSoundToPlay = sound;
	return self;
end;

function rockSpawnSoundEvent:readStream(streamId, connection)
	self.object = networkGetObject(streamReadInt32(streamId));
	self.rockSoundToPlay = streamReadInt32(streamId);
	self:run(connection);
end;

function rockSpawnSoundEvent:writeStream(streamId, connection)
	streamWriteInt32(streamId, networkGetObjectId(self.object));
	streamWriteInt32(streamId,self.rockSoundToPlay);
end;

function rockSpawnSoundEvent:run(connection)
	self.object:playSoundStatus(self.rockSoundToPlay, true);
	if not connection:getIsServer() then
		g_server:broadcastEvent(rockSpawnSoundEvent:new(self.object,self.rockSoundToPlay),nil,connection,self.object);
	end;
end;